package ru.dbotthepony.mc.otm.item;

import com.google.common.collect.ImmutableList;
import net.minecraft.ChatFormatting;
import net.minecraft.core.Direction;
import net.minecraft.nbt.*;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.TooltipFlag;
import net.minecraft.world.level.Level;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.matter.IPatternStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.capability.matter.PatternInsertStatus;
import ru.dbotthepony.mc.otm.capability.matter.PatternState;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;

public class ItemPatternStorage extends Item {
	public final int capacity;
	public boolean is_creative;

	public ItemPatternStorage(int capacity) {
		super(new Properties().tab(OverdriveThatMatters.CREATIVE_TAB).stacksTo(1));
		this.capacity = capacity;
		is_creative = false;
	}

	public ItemPatternStorage() {
		super(new Properties().tab(OverdriveThatMatters.CREATIVE_TAB).stacksTo(1));
		is_creative = true;
		capacity = Integer.MAX_VALUE;
	}

	@Nullable
	@Override
	public ICapabilityProvider initCapabilities(ItemStack stack, @Nullable CompoundTag nbt) {
		return new ItemPatternStorageCapability(stack, capacity);
	}

	@Override
	public void appendHoverText(ItemStack p_41421_, @Nullable Level p_41422_, List<Component> list, TooltipFlag p_41424_) {
		p_41421_.getCapability(MatteryCapability.PATTERN).ifPresent(capability -> {
			if (is_creative)
				list.add(new TranslatableComponent("otm.item.pattern.infinite.stored", capability.getStored()).withStyle(ChatFormatting.GRAY));
			else
				list.add(new TranslatableComponent("otm.item.pattern.stored", capability.getStored(), capability.getCapacity()).withStyle(ChatFormatting.GRAY));

			for (PatternState state : capability.getStoredPatterns()) {
				list.add(new TranslatableComponent("otm.item.pattern.line", state.item().getName(new ItemStack(state.item(), 1)), String.format("%.2f", state.research_percent() * 100d)).withStyle(ChatFormatting.AQUA));
			}
		});

		super.appendHoverText(p_41421_, p_41422_, list, p_41424_);
	}

	public static class ItemPatternStorageCapability implements IPatternStorage, ICapabilityProvider {
		private final LazyOptional<IPatternStorage> resolver = LazyOptional.of(() -> this);
		public final ItemStack stack;
		public final int capacity;

		public ItemPatternStorageCapability(ItemStack stack, int capacity) {
			this.stack = stack;
			this.capacity = capacity;
		}

		@Override
		public int getCapacity() {
			return capacity;
		}

		@Override
		public int getStored() {
			return getStoredPatterns().size();
		}

		@Nonnull
		@Override
		public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
			if (cap == MatteryCapability.PATTERN)
				return resolver.cast();

			return LazyOptional.empty();
		}

		@Nonnull
		@Override
		public Collection<PatternState> getStoredPatterns() {
			CompoundTag tag = stack.getOrCreateTag();

			if (!tag.contains("otm_patterns") || !(tag.get("otm_patterns") instanceof ListTag))
				return Collections.EMPTY_LIST;

			ListTag list = tag.getList("otm_patterns", Tag.TAG_COMPOUND);
			ArrayList<PatternState> state_list = new ArrayList<>();

			for (int i = 0; i < list.size(); i++) {
				var state = PatternState.deserializeNBT(list.getCompound(i));

				if (state != null)
					state_list.add(state);
			}

			return ImmutableList.copyOf(state_list);
		}

		@Override
		public PatternInsertStatus insertPattern(PatternState pattern, boolean only_update, boolean simulate) {
			CompoundTag tag = stack.getOrCreateTag();

			ListTag list;

			if (tag.get("otm_patterns") instanceof ListTag list1) {
				list = list1;
			} else {
				if (only_update)
					return new PatternInsertStatus();

				if (simulate) {
					if (capacity > 0)
						return new PatternInsertStatus(pattern);
					else
						return new PatternInsertStatus();
				}

				list = new ListTag();
				tag.put("otm_patterns", list);
			}

			String compare_with = Objects.requireNonNull(pattern.item().getRegistryName()).toString();

			int invalid_entries = 0;

			for (int i = 0; i < list.size(); i++) {
				var state = PatternState.deserializeNBT(list.getCompound(i));

				if (state != null) {
					if (state.equals(pattern)) {
						if (!simulate) {
							list.set(i, pattern.serializeNBT());
						}

						return new PatternInsertStatus(pattern, state);
					}
				} else {
					invalid_entries++;
				}
			}

			if (only_update || capacity <= list.size() - invalid_entries)
				return new PatternInsertStatus();

			if (invalid_entries > 0) {
				if (simulate)
					return new PatternInsertStatus(pattern);

				for (int i = 0; i < list.size(); i++) {
					var state = PatternState.deserializeNBT(list.getCompound(i));

					if (state == null) {
						list.set(i, pattern.serializeNBT());
						return new PatternInsertStatus(pattern);
					}
				}
			}

			if (!simulate)
				list.add(pattern.serializeNBT());

			return new PatternInsertStatus(pattern);
		}
	}
}
