package ru.dbotthepony.mc.otm.matter;

import com.google.common.collect.ImmutableList;
import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.item.Item;
import net.minecraft.world.level.Level;
import ru.dbotthepony.mc.otm.AbstractGrid;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.capability.matter.*;
import ru.dbotthepony.mc.otm.core.Fraction;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.*;
import java.util.function.Predicate;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class MatterGrid extends AbstractGrid<IMatterGridCell> implements IMatterGridListener {
	private final Set<IMatterGridListener> listeners = new HashSet<>();

	public static void scheduleDiscoverNeighbours(IMatterGridCell cell, BlockPos pos, Level level) {
		if (level instanceof ServerLevel)
			OverdriveThatMatters.tickUntil(level, () -> !cell.isValidMatterCell() || cell.connectOrCreateMatterGrid(pos, level, true));
	}

	public MatterGrid() {
		super(MatteryCapability.MATTER_CELL);
	}

	public Fraction getCapacity() {
		Fraction summ = Fraction.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				summ = summ.plus(handler.getMaxStoredMatter());
			}
		}

		return summ;
	}

	public Fraction getPotentialCapacity() {
		Fraction summ = Fraction.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				summ = summ.plus(handler.getMaxStoredMatter());
			}
		}

		return summ;
	}

	public Fraction getStored() {
		Fraction summ = Fraction.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				summ = summ.plus(handler.getStoredMatter());
			}
		}

		return summ;
	}

	public Fraction getPotentialStored() {
		Fraction summ = Fraction.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				summ = summ.plus(handler.getStoredMatter());
			}
		}

		return summ;
	}

	public Fraction extractMatter(Fraction howMuch, boolean simulate) {
		if (howMuch.compareTo(Fraction.ZERO) <= 0)
			return Fraction.ZERO;

		validate();

		Fraction extracted = Fraction.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				var drain = handler.extractMatterOuter(howMuch, simulate);

				if (!drain.equalsCompact(Fraction.ZERO)) {
					extracted = extracted.plus(drain);
					howMuch = howMuch.minus(drain);

					if (howMuch.compareTo(Fraction.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return extracted;
	}

	public Fraction softPushMatter(Fraction howMuch, boolean simulate) {
		if (howMuch.compareTo(Fraction.ZERO) <= 0)
			return Fraction.ZERO;

		validate();

		Fraction received = Fraction.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				var receive = handler.receiveMatterOuter(howMuch, simulate);

				if (!receive.equalsCompact(Fraction.ZERO)) {
					received = received.plus(receive);
					howMuch = howMuch.minus(receive);

					if (howMuch.compareTo(Fraction.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return received;
	}

	public Fraction pushMatter(Fraction howMuch, boolean simulate) {
		if (howMuch.compareTo(Fraction.ZERO) <= 0)
			return Fraction.ZERO;

		validate();

		Fraction received = Fraction.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.EXTRACT) {
				var receive = handler.receiveMatterOuter(howMuch, simulate);

				if (!receive.equalsCompact(Fraction.ZERO)) {
					received = received.plus(receive);
					howMuch = howMuch.minus(receive);

					if (howMuch.compareTo(Fraction.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return received;
	}

	public PatternInsertStatus insertPattern(PatternState state, boolean only_update, boolean simulate) {
		validate();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				var status = storage.insertPattern(state, true, simulate);

				if (status.status() != PatternInsertStatus.Status.FAIL) {
					return status;
				}
			}
		}

		if (only_update)
			return new PatternInsertStatus();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				var status = storage.insertPattern(state, false, simulate);

				if (status.status() != PatternInsertStatus.Status.FAIL) {
					return status;
				}
			}
		}

		return new PatternInsertStatus();
	}

	public Collection<PatternState> getStoredPatterns() {
		validate();

		ArrayList<PatternState> list = new ArrayList<>();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				list.addAll(storage.getStoredPatterns());
			}
		}

		return ImmutableList.copyOf(list);
	}

	public long getStoredPatternCount() {
		validate();

		long value = 0;

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				value += storage.getStored();
			}
		}

		return value;
	}

	public long getPatternCapacity() {
		validate();

		long value = 0;

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				value += storage.getCapacity();
			}
		}

		return value;
	}

	public Collection<PatternState> findPatterns(Item item) {
		return findPatterns((patternState -> item == patternState.item()));
	}

	public Collection<PatternState> findPatterns(Predicate<PatternState> item) {
		validate();

		ArrayList<PatternState> list = new ArrayList<>();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				list.addAll(storage.findPatterns(item));
			}
		}

		return ImmutableList.copyOf(list);
	}

	@Nullable
	public PatternState getPattern(PatternState state) {
		return getPattern(state.id());
	}

	@Nullable
	public PatternState getPattern(UUID id) {
		validate();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				var get = storage.getPattern(id);

				if (get != null) {
					return get;
				}
			}
		}

		return null;
	}

	public boolean hasPattern(UUID id) {
		return getPattern(id) != null;
	}

	@Nullable
	public MatterTaskAllocation allocateTask(boolean simulate) {
		validate();

		for (IMatterGridCell cell : cells) {
			IMatterTaskProvider tasks = cell.getTaskProvider();

			if (tasks != null) {
				MatterTaskAllocation allocated = tasks.allocateTask(simulate);

				if (allocated != null) {
					return allocated;
				}
			}
		}

		return null;
	}

	public boolean notifyTaskCompletion(MatterTask task) {
		validate();

		for (IMatterGridCell cell : cells) {
			IMatterTaskProvider tasks = cell.getTaskProvider();

			if (tasks != null) {
				if (tasks.notifyTaskCompletion(task)) {
					return true;
				}
			}
		}

		return false;
	}

	@Override
	protected void onAdded(IMatterGridCell cell) {
		cell.setMatterGrid(this);

		var tasks = cell.getTaskProvider();

		if (tasks != null) {
			for (var task : tasks.getAllTasks()) {
				onMatterTaskCreated(task);
			}
		}

		var patterns = cell.getPatternStorage();

		if (patterns != null) {
			for (var pattern : patterns.getStoredPatterns()) {
				onPatternAdded(pattern);
			}
		}
	}

	@Override
	protected void onRemoved(IMatterGridCell cell) {
		cell.setMatterGrid(null);

		var tasks = cell.getTaskProvider();

		if (tasks != null) {
			for (var task : tasks.getAllTasks()) {
				onMatterTaskRemoved(task);
			}
		}

		var patterns = cell.getPatternStorage();

		if (patterns != null) {
			for (var pattern : patterns.getStoredPatterns()) {
				onPatternRemoved(pattern);
			}
		}
	}

	public void attach(IMatterGridListener listener) {
		listeners.add(listener);
	}

	public void detach(IMatterGridListener listener) {
		listeners.remove(listener);
	}

	public void validate() {
		ArrayList<IMatterGridCell> invalid = new ArrayList<>();

		for (IMatterGridCell entity : cells) {
			if (!entity.isValidMatterCell()) {
				invalid.add(entity);
			}
		}

		if (invalid.size() != 0) {
			for (IMatterGridCell entity : invalid) {
				remove(entity);
			}
		}
	}

	@Override
	public void onPatternAdded(PatternState state) {
		// validate();
		cells.forEach(cell -> cell.onPatternAdded(state));
		listeners.forEach(cell -> cell.onPatternAdded(state));
	}

	@Override
	public void onPatternRemoved(PatternState state) {
		// validate();
		cells.forEach(cell -> cell.onPatternRemoved(state));
		listeners.forEach(cell -> cell.onPatternRemoved(state));
	}

	@Override
	public void onPatternUpdated(PatternState new_state, PatternState old_state) {
		// validate();
		cells.forEach(cell -> cell.onPatternUpdated(new_state, old_state));
		listeners.forEach(cell -> cell.onPatternUpdated(new_state, old_state));
	}

	@Override
	public void onMatterTaskCreated(MatterTask task) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskCreated(task));
		listeners.forEach(cell -> cell.onMatterTaskCreated(task));
	}

	@Override
	public void onMatterTaskUpdated(MatterTask new_state, MatterTask old_state) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskUpdated(new_state, old_state));
		listeners.forEach(cell -> cell.onMatterTaskUpdated(new_state, old_state));
	}

	@Override
	public void onMatterTaskFinished(MatterTask state) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskFinished(state));
		listeners.forEach(cell -> cell.onMatterTaskFinished(state));
	}

	@Override
	public void onMatterTaskRemoved(MatterTask state) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskRemoved(state));
		listeners.forEach(cell -> cell.onMatterTaskRemoved(state));
	}
}
