package ru.dbotthepony.mc.otm.core

import net.minecraft.nbt.ByteArrayTag
import net.minecraft.nbt.StringTag
import net.minecraft.nbt.Tag
import net.minecraft.network.FriendlyByteBuf
import java.math.BigDecimal
import java.math.BigInteger
import java.math.MathContext
import java.math.RoundingMode

fun powScale(int: Int): BigInteger {
	if (int <= 0)
		return BigInteger.ONE

	var result = BigInteger.TEN

	for (i in 2 .. int)
		result *= BigInteger.TEN

	return result
}

fun powUnscaled(unscaled: BigInteger, scale: Int): BigInteger {
	if (scale >= 0)
		return unscaled

	var result = unscaled

	for (i in 2 .. -scale)
		result *= BigInteger.TEN

	return result
}

val DEFAULT_MATH_CONTEXT = MathContext(64, RoundingMode.HALF_UP)

@Suppress("NOTHING_TO_INLINE")
inline fun invertCompare(int: Int): Int {
	if (int == 0)
		return 0

	return if (int < 0) 1 else -1
}

@JvmRecord
data class Fraction @JvmOverloads constructor(@JvmField val value: BigInteger, @JvmField val divisor: BigInteger = BigInteger.ONE, @JvmField val compact: Boolean = true) {
	@JvmOverloads constructor(value: Long, compact: Boolean = true) : this(BigInteger.valueOf(value), compact = compact)
	@JvmOverloads constructor(value: Int, compact: Boolean = true) : this(BigInteger.valueOf(value.toLong()), compact = compact)
	@JvmOverloads constructor(value: Float, compact: Boolean = true) : this(BigDecimal(value.toString()), compact = compact)
	@JvmOverloads constructor(value: Double, compact: Boolean = true) : this(BigDecimal(value.toString()), compact = compact)
	@JvmOverloads constructor(value: String, compact: Boolean = true) : this(BigDecimal(value), compact = compact)

	@JvmOverloads constructor(value: Long, div: Long, compact: Boolean = true) : this(BigInteger.valueOf(value), BigInteger.valueOf(div), compact = compact)
	@JvmOverloads constructor(value: Int, div: Int, compact: Boolean = true) : this(BigInteger.valueOf(value.toLong()), BigInteger.valueOf(div.toLong()), compact = compact)
	@JvmOverloads constructor(value: Float, div: Float, compact: Boolean = true) : this(BigDecimal(value.toString()), BigDecimal(div.toString()), compact = compact)
	@JvmOverloads constructor(value: Double, div: Double, compact: Boolean = true) : this(BigDecimal(value.toString()), BigDecimal(div.toString()), compact = compact)
	@JvmOverloads constructor(value: String, div: String, compact: Boolean = true) : this(BigDecimal(value), BigDecimal(div), compact = compact)

	@JvmOverloads constructor(value: BigDecimal, compact: Boolean = true) : this(powUnscaled(value.unscaledValue(), value.scale()), powScale(value.scale()), compact = compact)
	@JvmOverloads constructor(value: BigDecimal, div: BigDecimal, compact: Boolean = true) : this(powUnscaled(value.unscaledValue(), value.scale()).multiply(powScale(div.scale())), powScale(value.scale()).multiply(powUnscaled(div.unscaledValue(), div.scale())), compact = compact)

	fun compactAndCanonize(): Fraction {
		if (value == BigInteger.ZERO || value == BigInteger.ONE || divisor == BigInteger.ONE)
			return this

		val a = signnum()
		val b = value.signum()
		val c = divisor.signum()

		if (a != b && a != c) {
			val mod = value % divisor

			if (mod == BigInteger.ZERO)
				return Fraction(-value / -divisor, compact = compact)

			return Fraction(-value, -divisor)
		}

		val mod = value % divisor

		if (mod == BigInteger.ZERO)
			return Fraction(value / divisor, compact = compact)

		return this
	}

	fun isZero(): Boolean {
		return value == BigInteger.ZERO
	}

	fun compact(): Fraction {
		if (value == BigInteger.ZERO || value == BigInteger.ONE || divisor == BigInteger.ONE)
			return this

		val mod = value % divisor

		if (mod == BigInteger.ZERO)
			return Fraction(value / divisor, compact = compact)

		return this
	}

	fun canonize(): Fraction {
		if (value == BigInteger.ZERO || value == BigInteger.ONE || divisor == BigInteger.ONE || divisor == BigInteger.ZERO)
			return this

		val a = signnum()
		val b = value.signum()
		val c = divisor.signum()

		if (a != b && a != c)
			return Fraction(-value, -divisor)

		return this
	}

	// Операторы
	fun equalsCompact(other: Fraction?): Boolean {
		if (other == null)
			return false

		val a = compact()
		val b = other.compact()

		return a.value == b.value && a.divisor == b.divisor
	}

	operator fun compareTo(other: Fraction): Int {
		if (divisor == other.divisor)
			return value.compareTo(other.value)

		val a = signnum()
		val b = other.signnum()

		if (a == b && a == 0)
			return 0

		if (a < b)
			return -1
		else if (a > b)
			return 1

		val cmp = (value * other.divisor).compareTo(other.value * divisor)

		if (a != value.signum() && a != divisor.signum()) {
			if (b != other.value.signum() && b != other.divisor.signum()) {
				return cmp
			}

			return invertCompare(cmp)
		} else if (b != other.value.signum() && b != other.divisor.signum()) {
			return invertCompare(cmp)
		}

		return cmp
	}

	fun plusCompact(other: Fraction): Fraction {
		if (divisor == other.divisor) {
			if (divisor == BigInteger.ONE)
				return Fraction(value + other.value)

			val new = value + other.value

			if (new.compareTo(BigInteger.ZERO) == 0)
				return ZERO

			val mod = new % divisor

			if (mod == BigInteger.ZERO)
				return Fraction(new / divisor)

			return Fraction(new, divisor)
		}

		val new = value * other.divisor + other.value * divisor
		val div = divisor * other.divisor
		val mod = new % div

		if (mod == BigInteger.ZERO)
			return Fraction(new / div)

		return Fraction(new, div)
	}

	operator fun plus(other: Fraction): Fraction {
		if (compact)
			return plusCompact(other)

		if (divisor == other.divisor) {
			return Fraction(value + other.value, divisor, compact = false)
		}

		return Fraction(value * other.divisor + other.value * divisor, divisor * other.divisor, compact = false)
	}

	fun minusCompact(other: Fraction): Fraction {
		if (divisor == other.divisor) {
			if (divisor == BigInteger.ONE)
				return Fraction(value - other.value)

			val new = value - other.value

			if (new.compareTo(BigInteger.ZERO) == 0)
				return ZERO

			val mod = new % divisor

			if (mod == BigInteger.ZERO)
				return Fraction(new / divisor)

			return Fraction(new, divisor)
		}

		val new = value * other.divisor - other.value * divisor
		val div = divisor * other.divisor
		val mod = new % div

		if (mod == BigInteger.ZERO)
			return Fraction(new / div)

		return Fraction(new, div)
	}

	operator fun minus(other: Fraction): Fraction {
		if (compact)
			return minusCompact(other)

		if (divisor == other.divisor) {
			return Fraction(value - other.value, divisor, compact = false)
		}

		return Fraction(value * other.divisor - other.value * divisor, divisor * other.divisor, compact = false)
	}

	fun timesCompact(other: Fraction): Fraction {
		val new = value * other.value

		if (new.compareTo(BigInteger.ZERO) == 0)
			return ZERO

		val div = divisor * other.divisor
		val mod = new % div

		if (mod == BigInteger.ZERO)
			return Fraction(new / div)

		return Fraction(new, div)
	}

	operator fun times(other: Fraction): Fraction {
		if (compact)
			return timesCompact(other)

		return Fraction(value * other.value, divisor * other.divisor, compact = false)
	}

	fun divCompact(other: Fraction): Fraction {
		val new = value * other.divisor

		if (new.compareTo(BigInteger.ZERO) == 0)
			return ZERO

		val div = divisor * other.value
		val mod = new % div

		if (mod == BigInteger.ZERO)
			return Fraction(new / div)

		return Fraction(new, div)
	}

	operator fun div(other: Fraction): Fraction {
		if (compact)
			return divCompact(other)

		return Fraction(value * other.divisor, divisor * other.value, compact = false)
	}

	operator fun unaryMinus(): Fraction {
		return Fraction(-value, divisor)
	}

	operator fun unaryPlus(): Fraction {
		return this
	}

	operator fun plus(other: 	Float): Fraction = plus(Fraction(other))
	operator fun minus(other: 	Float): Fraction = minus(Fraction(other))
	operator fun times(other: 	Float): Fraction = times(Fraction(other))
	operator fun div(other: 	Float): Fraction = div(Fraction(other))

	operator fun plus(other: 	Double): Fraction = plus(Fraction(other))
	operator fun minus(other: 	Double): Fraction = minus(Fraction(other))
	operator fun times(other: 	Double): Fraction = times(Fraction(other))
	operator fun div(other: 	Double): Fraction = div(Fraction(other))

	// может вызвать путаницу
	/*
	operator fun plus(other: 	BigDecimal): Fraction = plus(Fraction(other))
	operator fun minus(other: 	BigDecimal): Fraction = minus(Fraction(other))
	operator fun times(other: 	BigDecimal): Fraction = times(Fraction(other))
	operator fun div(other: 	BigDecimal): Fraction = div(Fraction(other))

	operator fun plus(other: 	BigInteger): Fraction = plus(Fraction(other))
	operator fun minus(other: 	BigInteger): Fraction = minus(Fraction(other))
	operator fun times(other: 	BigInteger): Fraction = times(Fraction(other))
	operator fun div(other: 	BigInteger): Fraction = div(Fraction(other))
	 */

	operator fun plus(other: 	Int): Fraction = plus(Fraction(other))
	operator fun minus(other: 	Int): Fraction = minus(Fraction(other))
	operator fun times(other: 	Int): Fraction = times(Fraction(other))
	operator fun div(other: 	Int): Fraction = div(Fraction(other))

	operator fun plus(other: 	Long): Fraction = plus(Fraction(other))
	operator fun minus(other: 	Long): Fraction = minus(Fraction(other))
	operator fun times(other: 	Long): Fraction = times(Fraction(other))
	operator fun div(other: 	Long): Fraction = div(Fraction(other))

	/*
	fun add(other:				Float): Fraction = plus(Fraction(other))
	fun subtract(other:			Float): Fraction = minus(Fraction(other))
	fun multiply(other:			Float): Fraction = times(Fraction(other))
	fun divide(other:			Float): Fraction = div(Fraction(other))

	fun add(other:				Double): Fraction = plus(Fraction(other))
	fun subtract(other:			Double): Fraction = minus(Fraction(other))
	fun multiply(other:			Double): Fraction = times(Fraction(other))
	fun divide(other:			Double): Fraction = div(Fraction(other))

	fun add(other:				Int): Fraction = plus(Fraction(other))
	fun subtract(other:			Int): Fraction = minus(Fraction(other))
	fun multiply(other:			Int): Fraction = times(Fraction(other))
	fun divide(other:			Int): Fraction = div(Fraction(other))

	fun add(other:				Long): Fraction = plus(Fraction(other))
	fun subtract(other:			Long): Fraction = minus(Fraction(other))
	fun multiply(other:			Long): Fraction = times(Fraction(other))
	fun divide(other:			Long): Fraction = div(Fraction(other))

	fun add(other:				Fraction): Fraction = plus(other)
	fun subtract(other:			Fraction): Fraction = minus(other)
	fun multiply(other:			Fraction): Fraction = times(other)
	fun divide(other:			Fraction): Fraction = div(other)
	 */

	operator fun rem(other: Fraction): Fraction {
		return Fraction((this / other).wholePart())
	}

	// Преобразования
	fun toFloat(): Float {
		return (value / divisor).toFloat() + ((value % divisor).toFloat() / divisor.toFloat())
	}

	fun toDouble(): Double {
		return (value / divisor).toDouble() + ((value % divisor).toDouble() / divisor.toDouble())
	}

	fun toByteArray(): ByteArray {
		val bytesA = value.toByteArray()
		val bytesB = divisor.toByteArray()

		return byteArrayOf(bytesA.size.toByte(), *bytesA, bytesB.size.toByte(), *bytesB)
	}

	fun toInt(): Int {
		return (value / divisor).toInt()
	}

	fun toLong(): Long {
		return (value / divisor).toLong()
	}

	@JvmOverloads
	fun toBigDecimal(context: MathContext = DEFAULT_MATH_CONTEXT): BigDecimal {
		return BigDecimal(value).divide(BigDecimal(divisor), context)
	}

	// Утилиты
	fun wholePart(): BigInteger {
		return value / divisor
	}

	fun fractionPart(): Fraction {
		return Fraction(value % divisor, divisor)
	}

	fun modPart(): BigInteger {
		return value % divisor
	}

	override fun toString(): String {
		return "$value/$divisor"
	}

	fun formattedString(): String {
		return "${wholePart()} ${modPart()}/$divisor"
	}

	@JvmOverloads
	fun decimalString(nums: Int = 2, strict: Boolean = false): String {
		val whole = wholePart()
		val fraction = modPart()

		if (fraction == BigInteger.ZERO) {
			if (!strict)
				return whole.toString()

			return "${whole}.${"0".repeat(nums)}"
		}

		var strdec = (fraction.toDouble() / divisor.toDouble()).toString()

		if (strdec.length == 1) {
			strdec = ""

			if (strict) {
				strdec = "0".repeat(nums)
			}
		} else {
			strdec = strdec.substring(2)

			if (strict && strdec.length < nums) {
				strdec += "0".repeat(nums - strdec.length)
			} else if (strdec.length > nums) {
				strdec = strdec.substring(0, nums)
			}
		}

		if (strdec == "") {
			return whole.toString()
		}

		return "${whole}.$strdec"
	}

	fun signnum(): Int {
		val a = value.signum()
		val b = divisor.signum()

		if (a == b) {
			if (a == 0)
				return 0

			return 1
		}

		return -1
	}

	fun serializeNBT(): ByteArrayTag {
		return ByteArrayTag(toByteArray())
	}

	fun isNaN(): Boolean {
		return divisor == BigInteger.ZERO
	}

	fun write(buff: FriendlyByteBuf) {
		buff.writeByteArray(toByteArray())
	}

	fun max(vararg others: Fraction): Fraction {
		var max = this

		for (other in others) {
			if (max < other) {
				max = other
			}
		}

		return max
	}

	fun min(vararg others: Fraction): Fraction {
		var min = this

		for (other in others) {
			if (min > other) {
				min = other
			}
		}

		return min
	}

	fun moreThanZero(): Fraction {
		if (signnum() >= 0)
			return this

		return ZERO
	}

	fun lessOrZero(): Fraction {
		if (signnum() <= 0)
			return this

		return ZERO
	}

	companion object {
		@JvmField
		val ZERO = Fraction(BigInteger.ZERO)

		@JvmField
		val HALF = Fraction(BigInteger.ONE, BigInteger.TWO)

		@JvmField
		val THIRD = Fraction(BigInteger.ONE, BigInteger.valueOf(3))

		@JvmField
		val ONE = Fraction(BigInteger.ONE)

		@JvmField
		val TWO = Fraction(BigInteger.TWO)

		@JvmField
		val MINUS_ONE = Fraction(-1)

		@JvmField
		val TEN = Fraction(BigInteger.TEN)

		@JvmField
		val INT_MAX_VALUE = Fraction(Int.MAX_VALUE)

		@JvmField
		val LONG_MAX_VALUE = Fraction(Long.MAX_VALUE)

		@JvmStatic
		fun read(buff: FriendlyByteBuf): Fraction {
			return fromByteArray(buff.readByteArray())
		}

		@JvmStatic
		fun fromByteArray(bytes: ByteArray): Fraction {
			val bytesA = bytes.copyOfRange(1, 1 + bytes[0].toInt())
			val offsetB = 1 + bytes[0].toInt()
			val bytesB = bytes.copyOfRange(offsetB + 1, offsetB + 1 + bytes[offsetB].toInt())

			return Fraction(BigInteger(bytesA), BigInteger(bytesB))
		}

		@JvmStatic
		fun deserializeNBT(bytesTag: ByteArrayTag): Fraction {
			val bytes = bytesTag.asByteArray
			return fromByteArray(bytes)
		}

		// Преобразование строки в дробь с подавлением ошибок форматирования
		@JvmStatic
		fun fromString(str: String): Fraction {
			try {
				return Fraction(BigDecimal(str))
			} catch(err: Throwable) {
				return ZERO
			}
		}

		@JvmStatic
		fun deserializeNBT(tag: Tag?): Fraction {
			if (tag == null)
				return ZERO

			if (tag is ByteArrayTag)
				return deserializeNBT(tag)

			if (tag is StringTag)
				return try {
					fromString(tag.asString)
				} catch (anything: Throwable) {
					ZERO
				}

			return ZERO
		}
	}
}

